# This file is part of Grow-bots.
#
# Grow-bots is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Grow-bots is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Grow-bots.  If not, see <http://www.gnu.org/licenses/>.

import math
import random
import pygame
from OpenGL.GL import *
from OpenGL.GLU import *

import dyna
import load

"""
    Wow... What a mess...
"""

def deg_to_rad(deg):
    return deg * 0.0174532925

def render_cube():
    glBegin(GL_QUADS)
    glVertex3f(0.0, 1.0, 0.0)
    glVertex3f(1.0, 1.0, 0.0)
    glVertex3f(1.0, 1.0, 1.0)
    glVertex3f(0.0, 1.0, 1.0)

    glVertex3f(0.0, 0.0, 0.0)
    glVertex3f(1.0, 0.0, 0.0)
    glVertex3f(1.0, 1.0, 0.0)
    glVertex3f(0.0, 1.0, 0.0)

    glVertex3f(0.0, 0.0, 1.0)
    glVertex3f(1.0, 0.0, 1.0)
    glVertex3f(1.0, 1.0, 1.0)
    glVertex3f(0.0, 1.0, 1.0)

    glVertex3f(0.0, 0.0, 0.0)
    glVertex3f(0.0, 1.0, 0.0)
    glVertex3f(0.0, 1.0, 1.0)
    glVertex3f(0.0, 0.0, 1.0)

    glVertex3f(1.0, 0.0, 0.0)
    glVertex3f(1.0, 1.0, 0.0)
    glVertex3f(1.0, 1.0, 1.0)
    glVertex3f(1.0, 0.0, 1.0)
    glEnd()


class Editor:

    def __init__(self, game, size):
        self.game = game
        self.size = size
        self.cells = {}
        self.cursor = [size/2, size/2]
        self.dyn = self.game.dyn
        self.models = {
            "factories": load.OBJ("factory.obj"),
            "warehouses": load.OBJ("warehouse.obj"),
            "oilwells": load.OBJ("oilwell.obj"),
            "apartments": load.OBJ("apartment.obj"),
            "bootcamps": load.OBJ("bootcamp.obj"),
        }
        self.ground = load.texture("ground.png")
        self.gl_list = glGenLists(1)

    def recache(self):
        glNewList(self.gl_list, GL_COMPILE)
        glColor3ub(0xff, 0xff, 0xff)
        for cell, obj in self.cells.items():
            x, y = cell
            glPushMatrix()
            glTranslatef(x+0.5, 0.0, y+0.5)
            glCallList(self.models[obj].gl_list)
            glPopMatrix()
        glEndList()

    def destroy_half(self):
        half = int(len(self.cells) / 2)
        if half < 1:
            return []
        destroyed = []
        for cell in random.sample(self.cells, half):
            destroyed.append(self.cells[cell])
            del self.cells[cell]
        self.recache()
        return destroyed

    def destroy(self, percent):
        half = int(float(len(self.cells)) * percent)
        if half < 1:
            return []
        destroyed = []
        for cell in random.sample(self.cells, half):
            destroyed.append(self.cells[cell])
            del self.cells[cell]
        self.recache()
        return destroyed

    def build(self, building):
        cx, cy = map(int, self.cursor)
        if cx >= self.size:
            cx = self.size - 1
        if cy >= self.size:
            cy = self.size - 1

        if (cx, cy) in self.cells:
            return False

        self.cells[(cx, cy)] = building
        self.recache()
        return True

    def get_building(self):
        cx, cy = map(int, self.cursor)
        if cx >= self.size:
            cx = self.size - 1
        if cy >= self.size:
            cy = self.size - 1
        if (cx, cy) not in self.cells:
            return None
        return self.cells[(cx, cy)]

    def destroy_building(self):
        cx, cy = map(int, self.cursor)
        if cx >= self.size:
            cx = self.size - 1
        if cy >= self.size:
            cy = self.size - 1
        del self.cells[(cx, cy)]
        self.recache()

    def expand(self):
        self.size += 2
        new_cells = {}
        for cell, value in self.cells.items():
            new_cells[(cell[0]+1, cell[1]+1)] = value
        self.cells = new_cells
        self.cursor[0] += 1
        self.cursor[1] += 1
        self.recache()

    def move(self, dx, dy):
        rad = deg_to_rad(self.game.angle[1])
        s, c = math.sin(rad), math.cos(rad)
        self.cursor[0] += dx * c - dy * s
        self.cursor[1] += dx * s + dy * c
        for i in (0, 1):
            if self.cursor[i] < 0:
                self.cursor[i] = 0
            elif self.cursor[i] > self.size:
                self.cursor[i] = self.size

    def update(self):
        inkey = self.dyn.keys.is_down
        onkey = self.dyn.keys.on_down

        # Move faster when zoomed out (cover more ground)
        delta = (self.game.zoom + 2.0) / 70.0
        if delta > 3.0:
            delta = 3.0

        if inkey(pygame.K_d):
            self.move(delta * 0.9, 0.0)
        if inkey(pygame.K_a):
            self.move(-delta * 0.9, 0.0)
        if inkey(pygame.K_s):
            self.move(0.0, delta)
        if inkey(pygame.K_w):
            self.move(0.0, -delta)

    def render(self):
        half = float(self.size) / 2
        glPushMatrix()
        glDisable(GL_LIGHTING)
        glTranslatef(-self.cursor[0], 0.0,  -self.cursor[1])


        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, self.ground)
        glBegin(GL_QUADS)
        glColor3ub(0xff, 0xff, 0xff)
        glTexCoord2f(0.0, 0.0)
        glVertex3f(0.0, 0.0, 0.0)
        glTexCoord2f(self.size, 0.0)
        glVertex3f(self.size, 0.0, 0.0)
        glTexCoord2f(self.size, self.size)
        glVertex3f(self.size, 0.0, self.size)
        glTexCoord2f(0.0, self.size)
        glVertex3f(0.0, 0.0, self.size)
        glEnd()
        glDisable(GL_TEXTURE_2D)

        cx, cy = map(int, self.cursor)
        if cx >= self.size:
            cx = self.size - 1
        if cy >= self.size:
            cy = self.size - 1

        glPushMatrix()
        glBegin(GL_LINES)
        glColor3ub(0xff, 0x00, 0x00)
        glVertex3f(self.cursor[0], 1.0, self.cursor[1])
        glVertex3f(self.cursor[0], 0.0, self.cursor[1])
        glEnd()
        glTranslatef(cx, 0.0, cy)
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        glColor3ub(0x00, 0xff, 0x00)
        glEnable(GL_POLYGON_OFFSET_LINE)
        render_cube()
        glDisable(GL_POLYGON_OFFSET_LINE)
        glPopMatrix()
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glEnable(GL_LIGHTING)

        glCallList(self.gl_list)
        glPopMatrix()


def factory_cost(n):
    return 10 + (1.5 ** n) * 10

def warehouse_cost(n):
    return 50 + (2 ** n) * 100

def oilwell_cost(n):
    return 75 + (1.3 ** n) * 15

def bootcamp_cost(n):
    return 100 + (2 ** n) * 100

def apartment_cost(n):
    return 30 + (2 ** n) * 50

class Game:

    def __init__(self):
        dyna.init(width=640, height=480, fps=60)
        self.dyn = dyna.getcontext()
        self.editor = Editor(self, 4)
        self.angle = [30.0, 0.0]
        self.zoom = 5.0
        self.last_update = 0

        self.next_attack = pygame.time.get_ticks() + 5 * 60000
        self.attack_size = 100

        self.gui = dyna.Gui(
            pygame.font.Font(load.filepath("FreeMonoBold.ttf"), 13),
            (0xff, 0xff, 0xff),
            (0x33, 0x33, 0x33)
        )

        self.dyn.events.add("start", self.intro)
        self.dyn.events.add("render", self.render)
        self.dyn.events.add("update", self.update)
        self.dyn.events.add("update", self.editor.update)
        self.dyn.events.add("process", self.update_values)

        self.resources = {
            "junk": 500.0,
        }

        self.resources_max = {
            "junk": 0,
        }

        self.robots = {
            "boredbots": 500.0,
            "huntbots": 0.0,
            "armybots": 0.0,
        }

        self.robots_max = {
            "boredbots": 0.0,
            "huntbots": 0.0,
            "armybots": 0.0,
        }

        self.buildings = {
            "factories": 0,
            "warehouses": 0,
            "oilwells": 0,
            "apartments": 0,
            "bootcamps": 0,
        }

        self.costs = {}

        self.cost_func = {
            "oilwells": oilwell_cost,
            "factories": factory_cost,
            "apartments": apartment_cost,
            "warehouses": warehouse_cost,
            "bootcamps": bootcamp_cost,
        }

        self.recalc_costs()

        self.selected = "factories"

        glLightfv(GL_LIGHT0, GL_AMBIENT, (0.5, 0.5, 0.5, 1.0))
        glLightfv(GL_LIGHT0, GL_DIFFUSE, (1.0, 1.0, 1.0, 1.0))
        glEnable(GL_LIGHT0)
        glEnable(GL_LIGHTING)
        glEnable(GL_COLOR_MATERIAL)
        glEnable(GL_DEPTH_TEST)
        glShadeModel(GL_SMOOTH)
        glPolygonOffset(-2.0, 2)

    def start(self):
        self.dyn.start()

    def intro(self):
        self.pause(True)
        self.gui.alert([
            "Welcome to GrowBot!",
            "",
            "Beware of the humans...",
            "",
            "Press F1 for help",
        ])
        self.pause(False)

    def recalc_costs(self):
        for building in self.buildings:
            func = self.cost_func[building]
            self.costs[building] = func(self.buildings[building])

    def build(self):
        if self.resources["junk"] >= self.costs[self.selected]:
            if self.editor.build(self.selected):
                self.buildings[self.selected] += 1
                self.resources["junk"] -= self.costs[self.selected]
                self.recalc_costs()
        else:
            self.gui.alert("More junk required for '%s'" % self.selected)

    def pause(self, state):
        if state:
            self.dyn.events.remove("process", self.update_values)
        else:
            self.dyn.events.add("process", self.update_values)
            self.last_update = pygame.time.get_ticks()

    def destroy(self, percent):
        buildings = self.editor.destroy(percent)
        for building in buildings:
            self.buildings[building] -= 1
        return len(buildings)

    def bribe(self):
        half = int(self.resources["junk"] / 2)
        amount = random.randint(half, int(self.resources["junk"] + half))
        self.gui.alert([
            "Human:",
            "   So you want to make a deal, eh?",
            "   How about you give us " + str(amount) + " junk",
            "   and maybe we'll leave you alone... For now!",
        ])
        in_loop = True
        while in_loop:
            offer = int(self.gui.prompt("Your offer (junk): "))
            if offer <= self.resources["junk"]:
                in_loop = False
            else:
                self.gui.alert("You don't have that much junk!")
        r = random.randrange(8)

        if offer < amount:
            self.gui.alert([
                "Human:",
                "   This is what you offer us!?!?"
            ])
            if r < 2:
                self.gui.alert([
                    "Human:",
                    "   Ha, we don't need your junk anyway,",
                    "   foolish robots! But we shall take it",
                    "   and return another day.",
                ])
                self.resources["junk"] -= offer
            else:
                self.gui.alert([
                    "Human:",
                    "   You mock us with this offer, robot?",
                    "   Now you will pay!",
                    "   ...",
                ])
                buildings = self.destroy(0.5)
                self.resources["junk"] -= offer
                self.gui.alert([
                    "The humans attacked and "
                    "destroyed %i buildings!" % buildings
                ])
        elif offer == amount:
            if r < 2:
                self.gui.alert([
                    "Human:",
                    "   Thank you for the junk, robots!",
                    "   But we've decided to attack anyway!",
                    "   Ha ha ha ha ha...",
                    "   ...",
                ])
                buildings = self.destroy(0.3)
                self.resources["junk"] -= offer
                self.gui.alert([
                    "The humans attacked and "
                    "destroyed %i buildings!" % buildings
                ])
            else:
                self.gui.alert([
                    "Human:",
                    "   Alright robot, we will take this offer.",
                    "   But we will be back!",
                ])
                self.resources["junk"] -= offer
        else:
            if r == 1:
                self.gui.alert([
                    "Human:",
                    "   Stupid robot, we've tricked you!",
                    "   Now we will take your junk and attack!",
                    "   Ha ha ha ha ha...",
                    "   ...",
                ])
                buildings = self.destroy(0.25)
                self.resources["junk"] -= offer
                self.gui.alert([
                    "The humans attacked and "
                    "destroyed %i buildings!" % buildings
                ])
            else:
                self.gui.alert([
                    "Human:",
                    "   How generous of you, robot. We shall",
                    "   take this offer for now. But expect",
                    "   our return!",
                ])
                self.resources["junk"] -= offer

    def fight(self):
        r = random.randrange(4)
        if self.robots["armybots"] < self.attack_size:
            if r == 0:
                buildings = self.destroy(0.25)
            else:
                buildings = self.destroy(0.5)
            expanded = "No land has been obtained!"
        else:
            if r == 0:
                buildings = self.destroy(0.25)
            else:
                buildings = 0
            expanded = "Your land has been expanded!"
            self.editor.expand()
        self.gui.alert([
            "Battle report:",
            "   The humans have destoyed %i buildings!" % buildings,
            "   %s" % expanded,
        ])

    def attack(self):
        self.pause(True)
        self.next_attack = False
        self.gui.alert("Oh no! Humans are attacking!")
        in_loop = True
        while in_loop:
            a = self.gui.prompt("(f)ight or (b)ribe? ").lower()[0]
            if a == "f":
                self.fight()
                in_loop = False
            elif a == "b":
                self.bribe()
                in_loop = False
        self.pause(False)
        self.recalc_costs()
        self.next_attack = pygame.time.get_ticks() + 5 * 60000
        self.attack_size = self.robots["armybots"] + 100


    def update(self):

        inkey = self.dyn.keys.is_down
        onkey = self.dyn.keys.on_down

        speed = 3.0

        if inkey(pygame.K_UP):
            self.angle[0] -= speed
            if self.angle[0] < 0.0:
                self.angle[0] = 0.0
        if inkey(pygame.K_DOWN):
            self.angle[0] += speed
            if self.angle[0] > 90.0:
                self.angle[0] = 90.0
        if inkey(pygame.K_LEFT):
            self.angle[1] -= speed
        if inkey(pygame.K_RIGHT):
            self.angle[1] += speed
        if inkey(pygame.K_PAGEUP):
            self.zoom -= 0.5
            if self.zoom < 0.1:
                self.zoom = 0.1
        if inkey(pygame.K_PAGEDOWN):
            self.zoom += 0.5
            if self.zoom > float(self.editor.size) * 1.5:
                self.zoom = float(self.editor.size) * 1.5

        if onkey(pygame.K_1):
            self.selected = "factories"
        if onkey(pygame.K_2):
            self.selected = "oilwells"
        if onkey(pygame.K_3):
            self.selected = "warehouses"
        if onkey(pygame.K_4):
            self.selected = "apartments"
        if onkey(pygame.K_5):
            self.selected = "bootcamps"

        if onkey(pygame.K_SPACE):
            self.build()

        if onkey(pygame.K_DELETE):
            building = self.editor.get_building()
            if building:
                self.buildings[building] -= 1
                cost = self.cost_func[building](self.buildings[building])
                self.costs[building] = cost
                self.resources["junk"] += cost
                self.editor.destroy_building()

        if onkey(pygame.K_F1):
            self.gui.alert([
                "Keys:",
                "",
                "h = Hire robots (hire negative value to fire)",
                "",
                "1 = select Facorites",
                "2 = select Oil wells",
                "3 = select Warehouses",
                "4 = select Apartments",
                "5 = select Bootcamps",
                "",
                "Space = Build selected building at cursor",
                "Delete = Destroy selected building, get junk back!"
                "",
                "W, A, S, D = Move cursor",
                "Up, Left, Down, Right = Rotate view",
                "PageUp, PageDown = Zoom in / out",
                "",
                "F1 = Help screen :p",
                "Enter = Close popup window",
            ])

        if onkey(pygame.K_h):
            amount = self.gui.prompt("Amount to hire: ")
            try:
                amount = int(amount)
            except:
                self.gui.alert("Amount '%s' does not compute!" % amount)
                return
            robot = self.gui.prompt("Robot type: ").lower()
            if not robot.endswith("s"):
                robot += "s"
            if robot not in self.robots:
                self.gui.alert("Robot type '%s' does not compute!" % robot)
            else:
                if self.robots["boredbots"] >= amount:
                    self.robots["boredbots"] -= amount
                    self.robots[robot] += amount
                    self.gui.alert("%i %s hired!" % (amount, robot))
                else:
                    self.gui.alert("Does not compute! Need more bots!")

        if self.next_attack <= pygame.time.get_ticks():
            self.attack()

    def update_values(self):
        tick = pygame.time.get_ticks()
        if self.last_update != 0:
            delta = tick - self.last_update
        else:
            delta = 0.0
        self.last_update = tick
        delta /= 1000.0

        if self.robots["huntbots"] < 400:
            speed = (self.robots["huntbots"] / 125.0)
        elif self.robots["huntbots"] < 800:
            speed = (self.robots["huntbots"] / 150.0)
        elif self.robots["huntbots"] < 1600:
            speed = (self.robots["huntbots"] / 200.0)
        elif self.robots["huntbots"] < 3200:
            speed = (self.robots["huntbots"] / 225.0)
        else:
            speed = (self.robots["huntbots"] / 250.0)

        self.resources["junk"] += speed * delta

        self.resources_max["junk"] = self.costs["warehouses"]

        if self.resources["junk"] > self.resources_max["junk"]:
            self.resources["junk"] = self.resources_max["junk"]

        if self.buildings["factories"] < 4:
            speed = (float(self.buildings["factories"]) / 4.0)
        elif self.buildings["factories"] < 8:
            speed = (float(self.buildings["factories"]) / 5.0)
        elif self.buildings["factories"] < 16:
            speed = (float(self.buildings["factories"]) / 6.0)
        elif self.buildings["factories"] < 32:
            speed = (float(self.buildings["factories"]) / 7.0)
        else:
            speed = (float(self.buildings["factories"]) / 8.0)

        self.robots["boredbots"] += speed * delta

        self.robots_max["boredbots"] = 100 + self.buildings["apartments"] * 100
        if self.robots["boredbots"] > self.robots_max["boredbots"]:
            self.robots["boredbots"] = self.robots_max["boredbots"]

        self.robots_max["huntbots"] = 100 + self.buildings["oilwells"] * 200
        if self.robots["huntbots"] > self.robots_max["huntbots"]:
            self.robots["huntbots"] = self.robots_max["huntbots"]

        self.robots_max["armybots"] = self.buildings["bootcamps"] * 100
        if self.robots["armybots"] > self.robots_max["armybots"]:
            self.robots["armybots"] = self.robots_max["armybots"]

    def render(self):
        glTranslatef(0.0, -0.1, -self.zoom)
        glRotatef(self.angle[0], 1.0, 0.0, 0.0)
        glRotatef(self.angle[1], 0.0, 1.0, 0.0)
        cx, cy = map(lambda x: x - 0.5 * self.editor.size, self.editor.cursor)
        glLightfv(GL_LIGHT0, GL_POSITION,  (cx, 1.0, cy, 0.0))
        self.editor.render()
        self.dyn.push2d()

        y = 1
        for building, amount in self.buildings.items():
            self.gui.write(1, y, (building, ": ", int(amount)))
            y += 1

        y = 1
        for resource, amount in self.resources.items():
            self.gui.write(20, y, (
                resource, ": ",
                int(amount), "/", int(self.resources_max[resource])
            ))
            y += 1

        y = 1
        for robot, amount in self.robots.items():
            self.gui.write(40, y, (
                robot, ": ",
                int(amount), "/", int(self.robots_max[robot])
            ))
            y += 1

        self.gui.write(1, -2, [
            "Selected: ", self.selected,
            ", cost: ", int(self.costs[self.selected])
        ])

        if self.next_attack:
            diff = self.next_attack - pygame.time.get_ticks()
            attack_time = float(diff) / 60000
            measure = "minutes"
            if attack_time < 1.0:
                attack_time *= 60
                measure = "seconds"
            self.gui.write(1, -1, [
                "Next human attack: ", int(attack_time), " ", measure, ", ",
                int(self.attack_size), " humans!"
            ])
        self.dyn.pop2d()

if __name__ == "__main__":
    Game().start()
